#  R2P.rb
#  Ruby To Python
#  Devin Kendig
#  (c) 2006-09 ASGVIS, LLC

if( not file_loaded?("R2P.rb") )
puts "R2P.rb loaded"

require 'sketchup.rb'
require 'LangHandler.rb'
require 'vfs.rb'

require "rexml/document.rb"
include REXML  # so that we don't have to prefix everything with REXML::...

module VRayForSketchUp

    # 3 June 2008 - Joe
    # Only call this if you know what your doing...It will redirect all puts statements to that text file
    # and should really only be used in debug situations. Any call to redirect_stdout should have a matching
    # restore_stdout. Undefined behavior can happen if you call this multiple times without a matching restore_stdout call.
    # Really you should not call this more than once unless you mean to overwrite the previous dump file...notice that I open this
    # with mode "w" - that means we will wipe out the file each time we call this...
    def self.redirect_stdout()
        @orig_stdout = $stdout
        if isThisDarwin()
            filename = "/tmp/ruby_vfs_stdout.txt"
        else            
            filename = "#{getWindowsASGVISFolder()}/VfS/ruby_vfs_stdout.txt"
        end
            
        $stdout = File.new(filename, "w")                
    end
    
    # 3 June 2008 - Joe
    # See above notes...essentially the same
    def self.redirect_stderr()
        @orig_stderr = $stderr
        #$stderr = File.new("#{VRayForSketchUp.app_data_plugin_folder}stderr_debug.txt", "w" )
        if isThisDarwin()
            filename = "/tmp/ruby_vfs_stderr.txt"
        else
            filename = "#{getWindowsASGVISFolder()}/VfS/ruby_vfs_stderr.txt"
        end
            
        $stderr = File.new( filename, "w")          
    end
    
    # 3 June 2008 - Joe
    # Restores the stdout to what it was before I started mucking with it - read the above warnings
    def self.restore_stdout()
        if( defined?(@orig_stdout) and ( @orig_stdout != nil ) )
            $stdout = @orig_stdout
            @orig_stdout = nil
        end
    end
    
    # 3 June 2008 - Joe
    # See above notes...
    def self.restore_stderr()
        if( defined?(@orig_stderr) and ( @orig_stderr != nil ) )
            $stderr = @orig_stderr
            @orig_stderr = nil
        end        
    end    
end

# Set this to true if you want to redirect Sketchup stdout to a file
if false
    VRayForSketchUp.redirect_stdout()
    VRayForSketchUp.redirect_stderr()
end

if isThisDarwin()
  require 'RubyToPython/RubyToPython.bundle'
  require 'RubyToPython/VfSbin2xml.bundle'
else
  require "RubyToPython.so"
  require "VfSbin2xml.so"
end

require 'R2PCallbacks.rb'
require 'R2PRMsg.rb'
require 'VfSObservers.rb'
require 'skpHelperClasses.rb'
require 'VfSCommands.rb'
require 'VfSExport.rb'
require 'R2PToolbar.rb'

$vrayskpStrings = LanguageHandler.new("RubyToPython" )

PythonBridge.load_file("init")

def readHandler(msg)
    #puts "I just received a message from Python:"
    #puts msg
	msgParser(msg)
end# readHandler
PythonBridge.set_read_handler("readHandler")

# Example of how to send a properly formatted message to Python:
#argList = ["<material name='red'></material>"]
#evcb = "create_material"
#formattedMsg = fMsg(evcb,argList)
#PythonBridge.send(formattedMsg)
#
#Or you can do it all in one line like this:
#PythonBridge.send(fMsg("create_material",["<material name='red' color='red'></material>"]))
module VRayForSketchUp
    ## 5 May 2009 - Joe
    # These are just filler values for guids and other special marker items...they should be replaced with more valid values
    VFS_PRODUCT_GUID = "Something...Check the previous project to see the value we used"
    VFS_OPTIONS_DICTIONARY = "V-Ray for SketchUp Render Options Dictionary"
    VFS_MATERIAL_ATTRIBUTE = "V-Ray for SketchUp Material Attribute"
    VFS_LIGHT_ATTRIBUTE = "Check the old project"
    VFS_SPECIAL_OBJECT = "Check the old project ya putz-o"
    VFS_INFINITE_PLANE =  "Yeah - its an infinite plane"
    VFS_LIGHT_TABLE_ATTRIBUTE = "V-Ray for SketchUp Light Table"
    VFS_SUN = "V-Ray for SketchUp Sun Light"
    @@write_out_sun_xml = false
    @@hidden_lights_on = false
    
    def self.set_hidden_lights_on_val(newVal)
        @@hidden_lights_on = newVal
        
    end
    def self.get_hidden_lights_on_val()
        @@hidden_lights_on
    end
    
    def self.setupMenu()
        ourMenu = UI.menu("Plugins").add_submenu("V-Ray")
        lightMenu = ourMenu.add_submenu("Create Light")
        helpMenu = ourMenu.add_submenu("Help")
        debugMenu = helpMenu.add_submenu("Debugging")
        #Light Menu Items
        createOmniLight = UI::Command.new("Omni Light") {
            Sketchup.active_model.select_tool VRayForSketchUp::OmniLightTool.new
        }
        createRectLight = UI::Command.new("Rect Light") {
            Sketchup.active_model.select_tool VRayForSketchUp::RectLightTool.new
        }
        createSpotLight = UI::Command.new("Spot Light") {
            Sketchup.active_model.select_tool VRayForSketchUp::SpotLightTool.new
        }
        createIESLight = UI::Command.new("IES Light") {
            Sketchup.active_model.select_tool VRayForSketchUp::IESLightTool.new
        }
        lightMenu.add_item createOmniLight
        lightMenu.add_item createRectLight
        lightMenu.add_item createSpotLight
        lightMenu.add_item createIESLight
        #Help Menu Items
        generalHelp = UI::Command.new("Launch Help Page") {
            PythonBridge.send(fMsg("need_help",["General Support"]))
        }
        forumHelp = UI::Command.new("Launch VfS Forums") {
            PythonBridge.send(fMsg("need_help",["Forums"]))
        }
        trainingHelp = UI::Command.new("Launch Training Page") {
            PythonBridge.send(fMsg("need_help",["Get Training"]))
        }
        contactHelp = UI::Command.new("Contact Support") {
            PythonBridge.send(fMsg("need_help",["Contact Support"]))
        }
        helpMenu.add_item generalHelp
        helpMenu.add_item forumHelp
        helpMenu.add_item trainingHelp
        helpMenu.add_item contactHelp

        #Debug Menu Items
        killMats = UI::Command.new("Regenerate all V-Ray Materials") {
            regen_def_skp_mats()
        }
        syncMats = UI::Command.new("Force Sync for All Materials") {
            force_sync_mats()
        }
        clearTexCache = UI::Command.new("Clear Texture Cache Folder") {
            VRayForSketchUp.clear_texture_cache()
            UI.messagebox "Finished clearing texture cache"
        }
        debugMenu.add_item killMats
        debugMenu.add_item syncMats
        debugMenu.add_item clearTexCache
        #Main Menu Items
        showOptions = UI::Command.new("Options") {
            argList = ["whaddup!"]
            evcb = "show_optEd"
            formattedMsg = fMsg(evcb,argList)
            PythonBridge.send(formattedMsg)
        }
        showMatEd = UI::Command.new("Material Editor") {
            argList = ["whaddup!"]
            evcb = "show_matEd"
            formattedMsg = fMsg(evcb,argList)
            PythonBridge.send(formattedMsg)
        }
        showVFB = UI::Command.new("Show Frame Buffer") {
            argList = ["whaddup!"]
            evcb = "open_vfb"
            formattedMsg = fMsg(evcb,argList)
            PythonBridge.send(formattedMsg)
        }
        render = UI::Command.new("Render") {
            VfSExport.render
        }
        manualRender = UI::Command.new("Render From File") {
            PythonBridge.send(fMsg("start_manual_render",["render!!"]))
        }

        ourMenu.add_item showMatEd
        ourMenu.add_item showOptions
        ourMenu.add_item render
        ourMenu.add_item manualRender
        ourMenu.add_item showVFB

    end
    
    def self.getSunXML
       @@write_out_sun_xml
    end
    def self.setSunXML(newVal)
       @@write_out_sun_xml = newVal
    end
    
    def self.clear_vfs_scene_attribute()
      Sketchup.active_model.attribute_dictionaries.delete(VRayForSketchUp::VFS_PRODUCT_GUID)
    end
    # 6 May 2009 - Joe
    # This helper method will attach an attribute to the V-Ray for SketchUp attribute dictionary under the given name
    #
    # NOTE:: IMPORTANT IMPORTANT -- for some reason the attribute dictionary will not accept Hash instances
    # Arrays seem to work ok, as well as strings and numbers
    #
    def self.set_vfs_scene_attribute( name, value )
        #puts "got into set_vfs_scene_attribute"
        # Passing true as the second argument ensures that our VfS dictionary exists
        vfs_dictionary = Sketchup.active_model.attribute_dictionary(VRayForSketchUp::VFS_PRODUCT_GUID, true)
        vfs_dictionary[name] = value
        #puts "at the bottom of set_vfs_scene_attribute"
        return true
    end
    
    # 6 May 2009 - Joe
    # This helper will return the value stored for an attribute at the key "name". If none is there we return nil
    def self.get_vfs_scene_attribute( name )
        vfs_dictionary = Sketchup.active_model.attribute_dictionary(VRayForSketchUp::VFS_PRODUCT_GUID, false)
        return nil if vfs_dictionary == nil
        #puts "we got the vfs_dictionary and are going to return the value at this key"
        #val = vfs_dictionary[name]
        #puts "key is #{name.to_s} -- value is #{val.to_s}"
        return vfs_dictionary[name]
    end
    
    def self.load_existing_options()
        #UI.messagebox "in load_existing_options"
        options_dictionary = VRayForSketchUp.get_vfs_scene_attribute(VRayForSketchUp::VFS_OPTIONS_DICTIONARY)
        if options_dictionary == nil
            #UI.messagebox "why is this empty?"
            return nil
        end
        #for each option
        options_dictionary.each { |vfs_option| VRayForSketchUp.send_asset(vfs_option[1]) }
        #UI.messagebox "leaving load_existing_options"
        #if we got this far, we loaded everything sucessfully
        return 1
    end

    # 8 May 2009 - Joe
    # This helper function will attempt to attach userdata to a Sketchup material
    # @param[in] skp_mat - The Sketchup material to attach our userdata to    
    # @param[in] value - the value to assign to the vfs material attribute
    # @retval bool - true if we successfully apply the info to the attribute dictionary
    def self.set_vfs_material_attribute( skp_mat, value )   
        #UI.messagebox "in set_vfs_material_attribute, we're working with a material named: #{skp_mat.display_name}"
        return false if nil == skp_mat
        
        vfs_options_dictionary = skp_mat.attribute_dictionary(VRayForSketchUp::VFS_PRODUCT_GUID, true)
        if vfs_options_dictionary == nil
            #UI.messagebox "in set_vfs_material_attribute, unable to make or access our attribute dictionary on this material: #{skp_mat.display_name}"
            return false 
        end

        vfs_options_dictionary[VRayForSketchUp::VFS_MATERIAL_ATTRIBUTE] = value
        VRayForSketchUp.get_vfs_material_attribute(skp_mat)
        return true
    end

    # 8 May 2009 - Joe
    # This helper will check a Sketchup material for VfS material userdata
    # @param[in] skp_mat - the Sketchup material to check
    # @retval The information stored at our VfS Material attribute - or nil if nothing is found
    def self.get_vfs_material_attribute( skp_mat )
        return nil if nil == skp_mat
        #UI.messagebox "in get_vfs_material_attribute, so we're at least trying to find something in #{skp_mat.display_name}"
        vfs_dictionary = skp_mat.attribute_dictionary(VRayForSketchUp::VFS_PRODUCT_GUID, false)
        #UI.messagebox vfs_dictionary.to_s
        if nil == vfs_dictionary
            #UI.messagebox "yeah that was blank because we couldn't find an attribute_dictionary with our stuff attached (Y.Y)"
            return nil
        end

        return vfs_dictionary[VRayForSketchUp::VFS_MATERIAL_ATTRIBUTE]
    end
    
    # 8 May 2009 - Joe
    # This helper function will attempt to find a Sketchup scene material that matches the name provided
    # @param[in] name - a ruby string containing the name of a material to look for
    # @retval The sketchup material, or nil if we don't find one.
    def self.find_skp_material_by_name( name )
        # We have to do it by display_name because that is the name we're looking for...if we are attaching
        # to a built-in material like [Color_A01] then the display name is Color_A01 and it will not be found
        # if you look up Sketchup.active_model.materials["Color_A01"]
        0.upto(Sketchup.active_model.materials.length-1) do |index|
            mat = Sketchup.active_model.materials[index]
            if mat.display_name == name
                return mat
            end
        end
        return nil
    end
    
    def self.check_material_name( matName )
        if !(matName.include? "/") and (matName.length>0)
            #puts "in check_material_name, this name looks fine: [#{matName}]"
            return true
        end
   		#puts "in check_material_name, this name is invalid: [#{matName}]"
        return false
    end

    def self.fix_skp_material_name( originalMat, autoRename=1)
        #puts "in fix_skp_material_name"
        retMat = originalMat
        originalName = originalMat.display_name()
        isValid = VRayForSketchUp.check_material_name(originalName)
        if isValid == false
            newName = originalName.gsub("/", "\\")
            if autoRename == 1 and newName != originalName
                newName = VRayForSketchUp.hack_rename_skp_material( originalName, newName )
                oldMat = VRayForSketchUp.find_skp_material_by_name(originalName)
                newMat = VRayForSketchUp.find_skp_material_by_name(newName)

                newMat = VRayForSketchUp.find_skp_material_by_name(newName)
                if newMat != nil
                    retMat = newMat
                else
                    UI.messagebox "fix_skp_material_name failed while trying to rename #{originalName}"
                end
            end
        end
        #puts "leaving fix_skp_material_name"
        return retMat
    end#self.fix_skp_material_name

    #We need to make mats on our side
    def self.make_vray_mats_for_scene()
        #puts "in make_vray_mats_for_scene()"
        # for some reason the for loop style of looping was not always working...
        # for mat in Sketchup.active_model.materials was not working in some cases...
        #if you make a copy of the array of materials in the scene... the issue with .each goes away... hooray SketchUp!
        0.upto(Sketchup.active_model.materials.length-1) do |i|
            mat = Sketchup.active_model.materials[i]
            matOriginalName = mat.display_name
            #puts "processing #{matOriginalName}..."
            validName = VRayForSketchUp.check_material_name(mat.display_name)
            if validName == false
                #puts "about to fix a material name: #{mat.display_name}, mat count is #{Sketchup.active_model.materials.count}"
                mat = VRayForSketchUp.fix_skp_material_name(mat)
                #puts "just fixed a material name: #{mat.display_name}, mat count is #{Sketchup.active_model.materials.count}"
            end

        	if mat == nil or mat.deleted?
        	   puts "in make_vray_mats_for_scene(), current mat is nil or deleted: #{mat}"
        	   return
        	end
            VRayForSketchUp.hack_ensure_texture_exists(mat)
            if VRayForSketchUp.get_vfs_material_attribute(mat)
                #this material already has our stuff attached to it
                #UI.messagebox "in make_vray_mats_for_scene, found existing user data for #{mat.display_name}"
                skpMatXML = VRayForSketchUp.get_vfs_material_attribute(mat)
                skpMatXML = VRayForSketchUp.checkNameInXml(mat.display_name(), skpMatXML)
                if skpMatXML == nil or matOriginalName != mat.display_name
                    searchURL = mat.display_name()
                    #puts "Error: could not retrieve serialized data for material named: #{mat.display_name}, using defaults."
                    CallbackWatcher.startCallback("make_vray_mats_for_scene")
                    skpMat = SketchupMaterial.new()
                    skpMat.fromSkp(mat)
                    #puts "in make_vray_mats_for_scene, found mat attribute, current skpMat.name = #{skpMat.name}"
                    skpMatXML = skpMat.toXML()
                    matList = [skpMatXML]
                    PythonBridge.send(fMsg("create_material",matList))
                    CallbackWatcher.lastMaterialXML = skpMatXML
                    CallbackWatcher.endCallback()
                    next
                end
                VRayForSketchUp.send_asset(skpMatXML)


                #VRayForSketchUp.send_asset(skpMatXML)
            else
                skpMat = SketchupMaterial.new()
                skpMat.fromSkp(mat)
                skpMatXML = skpMat.toXML()
                matList = [skpMatXML]
                
                # CallbackWatcher can protect us from re-entering this
                CallbackWatcher.startCallback("create_material")
                PythonBridge.send(fMsg("create_material",matList))
                CallbackWatcher.endCallback()
            end
        end
        #puts "leaving make_vray_mats_for_scene()"
    end
    
    def self.quietly_force_sync_mats()
        #puts "in quietly_force_sync_mats"
        matList = []
        0.upto(Sketchup.active_model.materials.length-1) do |i|
            mat = Sketchup.active_model.materials[i]
            matList.push mat.display_name if VRayForSketchUp.check_material_name(mat.display_name) != false
        end
        PythonBridge.send(fMsg("force_sync_mats", matList)) if matList.length>0
    end#self.make_vray_mats_for_compdef()
    
    def self.delayed_force_sync_mats(delay)
        #puts "in delayed_force_sync_mats"
        PythonBridge.send(fMsg("delayed_force_sync_mats", [delay]))
    end

    def self.regen_vray_mats_for_scene()
        # for some reason the for loop style of looping was not always working...
        # for mat in Sketchup.active_model.materials was not working in some cases...
        Sketchup.active_model.materials.count.times do |index|
            mat = Sketchup.active_model.materials[index]
            VRayForSketchUp.hack_ensure_texture_exists(mat)
            if VRayForSketchUp.get_vfs_material_attribute(mat)
                #this material already has our stuff attached to it
                #UI.messagebox "in make_vray_mats_for_scene, found existing user data for #{mat.display_name}"
                mat.delete_attribute VRayForSketchUp::VFS_PRODUCT_GUID
            end
            skpMat = SketchupMaterial.new()
            skpMat.fromSkp(mat)
            skpMatXML = skpMat.toXML()
            matList = [skpMatXML]
            # CallbackWatcher can protect us from re-entering this
            CallbackWatcher.startCallback("create_material")
            PythonBridge.send(fMsg("create_material",matList))
            CallbackWatcher.endCallback()
        end
    end
    
    # 13 May 2009 - Devin
    # This is a helper method to attach a value to the entity in our infinate plane attribute dictionary
    def self.set_vfs_plane_attribute( entity, value )
        #print "in R2P.rb, line 135. def self.set_vfs_plane_attribute( entity, value ). Here is the contents of value:"
        #print "----"
        #print value
        #print "----"
        return false if nil == entity

        vfs_dictionary = entity.attribute_dictionary(VRayForSketchUp::VFS_PRODUCT_GUID, true)
        return false if vfs_dictionary == nil
        
        vfs_dictionary[VRayForSketchUp::VFS_INFINITE_PLANE] = value
        return true
    end

    # 13 May 2009 - Devin
    # This is just a helper method to extract the infinate plane information ( if any exists ) on the given entity
    def self.get_vfs_plane_attribute( entity )
        return nil if nil == entity
                
        # We'll try to extract our product attribute dictionary...we will not create one if it does not exist
        vfs_dictionary = entity.attribute_dictionary(VRayForSketchUp::VFS_PRODUCT_GUID, false)
        return nil if vfs_dictionary == nil
        return vfs_dictionary[VRayForSketchUp::VFS_INFINITE_PLANE]
    end

    # 9 May 2009 - Joe
    # This is a helper method to attach a value to the entity in our light attribute dictionary
    def self.set_vfs_light_attribute( entity, value )
        return false if nil == entity
        
        vfs_dictionary = entity.attribute_dictionary(VRayForSketchUp::VFS_PRODUCT_GUID, true)
        return false if vfs_dictionary == nil
        
        vfs_dictionary[VRayForSketchUp::VFS_LIGHT_ATTRIBUTE] = value
        return true
    end
    
    # 9 May 2009 - Joe
    # This is just a helper method to extract the light information ( if any exists ) on the given entity
    def self.get_vfs_light_attribute( entity )
        return nil if nil == entity
                
        # We'll try to extract our product attribute dictionary...we will not create one if it does not exist
        vfs_dictionary = entity.attribute_dictionary(VRayForSketchUp::VFS_PRODUCT_GUID, false)
        return nil if vfs_dictionary == nil
        
        return vfs_dictionary[VRayForSketchUp::VFS_LIGHT_ATTRIBUTE] 
    end
       
    # 4 Aug 2009 - Joe
    # This will iterate our light table and send the assets over
    # 11 Nov 2009 - Joe
    # Since we no longer use the light table we will iterate over the component definitions table and 
    # add any lights we find. Since we use groups for lights they will be in the compdef table
    def self.load_existing_lights()
        for comp in Sketchup.active_model.definitions
            for inst in comp.instances
               xml = self.get_vfs_light_attribute( inst )
               VRayForSketchUp.send_asset( xml ) if ( xml != nil )
            end
        end
    end    
    
    # 4 June 2009 - Justin
    # This is a helper method to attach a value to the entity in our sun attribute dictionary
    def self.set_vfs_sun_attribute( value )
        vfs_dictionary = Sketchup.active_model.attribute_dictionary(VRayForSketchUp::VFS_PRODUCT_GUID, true)
        return false if vfs_dictionary == nil
        
        vfs_dictionary[VRayForSketchUp::VFS_SUN] = value
        return true
    end
    
    # 4 June 2009 - Justin
    # This is just a helper method to extract the sun information ( if any exists ) on the given entity
    def self.get_vfs_sun_attribute(  )
        # We'll try to extract our product attribute dictionary...we will not create one if it does not exist
        vfs_dictionary = Sketchup.active_model.attribute_dictionary(VRayForSketchUp::VFS_PRODUCT_GUID, false)
        return nil if vfs_dictionary == nil
        
        return vfs_dictionary[VRayForSketchUp::VFS_SUN]
    end
    
    # 4 June 2009 - Justin
    # Check to see if the current model has a sun
    def self.has_sun( )
        return nil != VRayForSketchUp.get_vfs_sun_attribute( )
    end

    # 9 May 2009 - Joe
    # Check to see if this object is holding vfs light information
    def self.is_light( entity )
        return false if ( nil == entity )
        return nil != VRayForSketchUp.get_vfs_light_attribute( entity )
    end

    # 9 May 2009 - Joe
    # This will handle editing the selected light - it must be in Sketchup.active_model.selection[0]
    def self.edit_light( )
        entity = Sketchup.active_model.selection[0]
        return false if ( not VRayForSketchUp.is_light( entity ) )
        
        # 4 Aug 2009 - Joe
        # When we added the light table we stopped storing the xml on the light object - we
        # just put the url on that, and then lookup the xml in our table using that url
        light_xml = VRayForSketchUp.get_vfs_light_attribute( entity )
        return false if( nil == light_xml )

        # This will launch our editor modally
        #puts "About to call edit_light with #{light_xml}"
        message = fMsg("edit_light", [light_xml] )
        PythonBridge.send( message )
    end
    
    # 13 May 2009 - Devin
    # Check to see if this object is holding vfs light information
    def self.is_plane( entity )
        return false if ( nil == entity )
        return nil != VRayForSketchUp.get_vfs_plane_attribute( entity )
    end
    
    def self.round4Down(num)
        # equivalent to num - ( num % 4 )
        return num - num.modulo( 4 )
    end
    
    def self.round4Up(num)
        down = self.round4Down(num)
        # if we're not already at an even divisor ( which is the only way num == down...) then we add 4
        return down + 4 if( num != down ) 
            
        return down
    end

    # 10 May 2009 - Joe
    # In our old version we made a V-Ray Sphere then converted it to Skp faces...Maybe we can make it using the follow tool instead?
    def self.create_vray_sphere( radius, subdivs, tm )
        sphereOrigin = Geom::Point3d.new
        
        # 2 Nov 2009 - Joe, Devin, Justin
        # Apparently we get random crashes when the subdiv is not a multiple of 4...this reminds me of
        # the memset / memcpy problems I had in the vfb when the pixel count was not a multiple of 4, so perhaps
        # Sketchup does some kind of memcpy/memset using the subdivs or geometry made from the subdivs. So we
        # auto-correct the desired subdiv count to the next multiple of 4. This seems to fix the issue...who knows...
        subdivs = self.round4Up( subdivs )

        model = Sketchup.active_model
        model.start_operation("Draw Sphere")
        
        # 28 May 2009 - Joe
        # Thomas Thomassen pointed out that we need to add to active_entities instead of entities
        # or we will be adding things to the incorrect place in some situations
        group = model.active_entities.add_group
        entities = group.entities
          
        circle1 = entities.add_circle sphereOrigin, [0,1,0], radius, subdivs
        circle2 = entities.add_circle sphereOrigin+[0,0,-2*radius], [0,0,1], 2.0 * radius, subdivs
        face1 = entities.add_face circle1
        #face2 = entities.add_face circle2
        finalSphere = face1.followme circle2
        entities.erase_entities circle2
        model.commit_operation

        #circle = group.entities.add_circle( Geom::Point3d.new(0.0,0.0,0.0), Geom::Vector3d.new(0.0,0.0,1.0), radius, subdivs )
        group.transform! tm
        return group
    end
    
    # 11 May 2009 - Joe
    # This will check to see if the provided Sketchup material has V-Ray Material Userdata on it
    def self.is_vray_material( material )
        return false if( nil == material )
        return nil != VRayForSketchUp.get_vfs_material_attribute( material )        
    end
    
    # This was inspired by http://www.ruby-doc.org/core/classes/Dir.src/M002324.html
    # But I took out hte module variable references and the $SAFE stuff
    def self.get_tempdir
        tmp = '.'
        
        for dir in [ENV['TMPDIR'], ENV['TMP'], ENV['TEMP'], ENV['USERPROFILE'], '/tmp']
          if dir and File.directory?(dir) and File.writable?(dir)
              tmp = dir
              break
          end
        end
        
        return File.expand_path(tmp)      
    end
    
    def self.texture_cache_folder
        folder = File.join( self.get_tempdir, "VfSTextureCache" )
               
        if !File.exists?( folder )
           Dir.mkdir( folder ) 
        end
        return folder
    end

    def self.clear_texture_cache
          folder = self.texture_cache_folder
          # call the python function empty_folder passing it folder
          PythonBridge.send(fMsg("empty_folder", [folder]))
          # We do this in python now to account for unicode file names
          # remove any files in that folder to avoid too much stuff accumulating
          #dir = Dir.new( folder )
          
          #for name in dir.to_a
          #    puts name
          #    fullPath = File.join( dir.path, name )
          #    if File.file?( fullPath )
          #        print "the file is "
          #        print fullPath
          #        print "/n"
          #        File.delete( fullPath )
          #    end
          #end
    end
            
	class << self
		# 30 Jan 2008 - Joe
		# add attributes for our observers if this is sketchup 6 - before I was actually writing out the reader/writers...
		attr_accessor( :app_observer, :material_observer, :model_observer, :definitions_observer) if( Sketchup.version_number >= 6000000 )
	end #class << self
end # module VRayForSketchUp
#now let's set up our observer stuff
if( Sketchup.version_number >= 6000000 )
	VRayForSketchUp.app_observer= VRayForSketchUp::VfSAppObserver.new
	VRayForSketchUp.model_observer= VRayForSketchUp::VfSModelObserver.new
	VRayForSketchUp.definitions_observer= VRayForSketchUp::VfSDefinitionsObserver.new
	VRayForSketchUp.material_observer= VRayForSketchUp::VfSMaterialsObserver.new
	# add our application observer to Sketchup
	Sketchup.add_observer( VRayForSketchUp.app_observer )
	#set up our initial V-Ray menu in the Plugins menu
	VRayForSketchUp.setupMenu()
		
if( Sketchup.active_model )
    # 16 Apr 2008 - Joe  ( Case 381 )
    # If a user double-clicks on a skp file to open Sketchup then we can not simply 
    # call the handle_new_model function, it requires an onOpenModel event to properly load our info
    # so I will do that instead...
    #VRayForSketchUp.handle_new_model( Sketchup.active_model ) 
    VRayForSketchUp.app_observer.onOpenModel( Sketchup.active_model )
end # if( Sketchup.active_model )
	
end # if( Sketchup.version_number >= 6000000 )

end # if( not file_loaded?("R2P.rb") )
file_loaded("R2P.rb")